using System;
using System.Windows.Forms;
using System.Threading;
using System.Timers;
using System.IO;
using NationalInstruments.DAQmx;

namespace accel_compass
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	public class Robot
	{
		private string leftMotorDevice = "Dev1/ao0";
		private string rightMotorDevice = "Dev1/ao1";
		private string accelDevice1 = "Dev1/ai2";
		private string accelDevice2 = "Dev1/ai3";
		private string compassDevice = "Dev1/ai7";

		private double leftMotor = 0;
		private double rightMotor = 0;


		private double accelDev1_calibValue;
		private double accelDev2_calibValue;
		private double accelDev1_actualValue;
		private double accelDev2_actualValue;

		private double compassDev_actualValue;

		private bool digitalCompassActive = false;
		private double digitalCompassHeading = 0;
		private ThreadStart digitalCompassThreadStart;
		private Thread digitalCompassThread;
		
		private HeadingFollowerBehavior hfBehavior = null;
		private BumpDetectBehavior bdBehavior = null;
		private int bumpCount = 0;

		private double x_W, y_W, theta_W;
		private double vx, vy; 
		
		//referencial: vx = -ai2      vy = ai3


		private Task taskAccel;
		private AnalogMultiChannelReader readerAccel;

		private Task taskCompass;
		private AnalogMultiChannelReader readerCompass;

		private double sampleRateAccel = 5000;
		private int samplesPerStepAccel = 500; //so we will run calculations routine 10 times per second

		private double sampleRateCompass = 5000;
		private int samplesPerStepCompass = 2500;

		

		private double k_voltageToAccel = 68;


		private static Robot _currentInstance = null;

		public static Robot getInstance()
		{
			if(_currentInstance == null) 
			{
				_currentInstance = new Robot();
			}

			return _currentInstance;

		}

		private Robot()
		{
			accelDev1_calibValue = 0;
			accelDev2_calibValue = 0;
			
			resetOdometry();

			initAccelDevices();
			initCompassDevice();

			digitalCompassThreadStart = new ThreadStart(digitalCompassThreadFunc);
			digitalCompassThread = new Thread(digitalCompassThreadStart);
			digitalCompassThread.Start();
			
			hfBehavior = new HeadingFollowerBehavior(this);
			bdBehavior = new BumpDetectBehavior(this);

			this.stop();


		} // Robot()

		public void Halt()
		{
			this.stop();
			if(digitalCompassThread != null)
			{
				digitalCompassThread.Abort();
			}
			hfBehavior.terminate();
			bdBehavior.terminate();
		}

		public void setLeftMotor(double value) 
		{
			leftMotor = value;
			
			// task for left motor
			try
			{
				using (Task leftTask = new Task())
				{
					leftTask.AOChannels.CreateVoltageChannel(leftMotorDevice, "aoChannel",
						0, 5,
						AOVoltageUnits.Volts);
					AnalogSingleChannelWriter leftWriter = new AnalogSingleChannelWriter(leftTask.Stream);

					leftWriter.WriteSingleSample(true, value); // stop motor
				}
			}
			catch (DaqException ex)
			{
				MessageBox.Show(ex.Message);
			}
		} // setLeftMotor

		public void setRightMotor(double value)
		{
			rightMotor = value;
			
			// task for right motor
			try
			{
				using (Task rightTask = new Task())
				{
					rightTask.AOChannels.CreateVoltageChannel(rightMotorDevice, "aoChannel",
						0, 5,
						AOVoltageUnits.Volts);
					AnalogSingleChannelWriter rightWriter = new AnalogSingleChannelWriter(rightTask.Stream);

					rightWriter.WriteSingleSample(true, value); // stop motor
				}
			}
			catch (DaqException ex)
			{
				MessageBox.Show(ex.Message);
			}
		} // setRightMotor

		public void stop()
		{
			this.setLeftMotor(0);
			this.setRightMotor(0);
		} // stop


		private void initAccelDevices()
		{

			taskAccel = new Task("Accel Task");


			//Create a virtual channel
			try
			{
				taskAccel.AIChannels.CreateVoltageChannel(getAccelDevice1() ,"",
				AITerminalConfiguration.Rse,0,
				5,AIVoltageUnits.Volts);  



				taskAccel.AIChannels.CreateVoltageChannel(getAccelDevice2() ,"",
				AITerminalConfiguration.Rse,0,
				5,AIVoltageUnits.Volts);  


				// Configure timing specs    
				taskAccel.Timing.ConfigureSampleClock("", sampleRateAccel, SampleClockActiveEdge.Rising,
					SampleQuantityMode.FiniteSamples, samplesPerStepAccel);


				readerAccel = new AnalogMultiChannelReader(taskAccel.Stream);

				//Verify the Task

				taskAccel.Control(TaskAction.Verify);
			}
			catch (DaqException ex)
			{
				//MessageBox.Show(ex.Message);
				//taskAccel.Dispose();
			}

		}

		private void initCompassDevice()
		{

			taskCompass = new Task("Compass Task");


			//Create a virtual channel
			try
			{
				taskCompass.AIChannels.CreateVoltageChannel(getCompassDevice() ,"",
					AITerminalConfiguration.Rse,0,
					5,AIVoltageUnits.Volts);  


				// Configure timing specs    
				taskCompass.Timing.ConfigureSampleClock("", sampleRateCompass, SampleClockActiveEdge.Rising,
					SampleQuantityMode.FiniteSamples, samplesPerStepCompass);

				//Verify the Task
				taskCompass.Control(TaskAction.Verify);


				readerCompass = new AnalogMultiChannelReader(taskCompass.Stream);

			}
			catch (DaqException ex)
			{
				MessageBox.Show(ex.Message);
				taskCompass.Dispose();
			}
		}

		private double [,] getMultiSampleOnReader(AnalogMultiChannelReader reader, int samplesPerStep)
		{
			return reader.ReadMultiSample(samplesPerStep);
		}



			//this step must be called 10 times per second!!!
		public void odometryStep()
		{

			double [,] samples = getMultiSampleOnReader(readerAccel, samplesPerStepAccel);

			for(int i = 0; i < samples.GetLength(1); i++)
			{	
				accelDev1_actualValue = convertVoltageToAccel1( samples[0, i] );
				accelDev2_actualValue = - convertVoltageToAccel2( samples[1, i] );

				//accelDev1_actualValue = Math.Round(accelDev1_actualValue, 2);
				//accelDev2_actualValue = Math.Round(accelDev2_actualValue, 2);

				//System.Console.WriteLine("Accel1 = " + accel1.ToString() );
				//System.Console.WriteLine("Accel2 = " + accel2.ToString() );
				vx = vx + accelDev1_actualValue * (1/sampleRateAccel);
				vy = vy + accelDev2_actualValue * (1/sampleRateAccel);

				//quando tivermos a bussula temos que fazer uma transformao aqui!! para passarmos do referencial do robot para o referencial do mundo
				x_W = x_W + vx * (1/sampleRateAccel);
				y_W = y_W + vy * (1/sampleRateAccel);

			}


		}


		public void resetOdometry()
		{
			x_W = 0;
			y_W = 0;
			theta_W = 0;
			vx = 0;
			vy = 0;

		}

		public void compassRead()
		{
			double [,] samples = getMultiSampleOnReader(readerCompass, samplesPerStepCompass);
			
			double [] sampleVector = new double[samples.GetLength(1)];

			for(int i = 0; i < samples.GetLength(1); i++)
			{	
				sampleVector[i] = samples[0,i];
			}
			
			compassDev_actualValue = convertVoltageToHeading(Stat.GetAverage(sampleVector));

		}


		public void digitalCompassStart()
		{
			digitalCompassActive = true;
		}

		public void digitalCompassStop()
		{
			digitalCompassActive = false;
		}

		private void digitalCompassThreadFunc()
		{

			/*Task digitalReadTask;
			DigitalSingleChannelReader digitalReader;*/

			Task taskDigitalCompass = new Task("Digital Compass Task");



			int sampleRate = 10000;
			int numSamples = (int)(Math.Ceiling(((double)sampleRate) * 0.102) * 2);
			AnalogMultiChannelReader readerDigitalCompass = null;


			bool writeToFile = true;
			TextWriter tw = new StreamWriter("C:\\digitalCompassSignal.txt");

			try
			{



				taskDigitalCompass.AIChannels.CreateVoltageChannel(getCompassDevice() ,"",
					AITerminalConfiguration.Rse,0,
					5,AIVoltageUnits.Volts);  


				// Configure timing specs    
				taskDigitalCompass.Timing.ConfigureSampleClock("", sampleRate, SampleClockActiveEdge.Rising,
					SampleQuantityMode.FiniteSamples, numSamples);

				//Verify the Task
				taskDigitalCompass.Control(TaskAction.Verify);


				readerDigitalCompass = new AnalogMultiChannelReader(taskDigitalCompass.Stream);



/*				digitalReadTask = new Task("DigitalCompass");

				digitalReadTask.DIChannels.CreateChannel(
					"Dev1/port0/line0:7",
					"myChannel",
					ChannelLineGrouping.OneChannelForAllLines);

				digitalReader = new DigitalSingleChannelReader(digitalReadTask.Stream);*/

			}
			catch(DaqException exception)
			{
				MessageBox.Show(exception.Message);
				//digitalReader = null;
				//dispose task
				//digitalReadTask.Dispose();
			}


			//bool lineHigh = false;

			long startTime = System.Environment.TickCount;
			long endTime = System.Environment.TickCount;

			while(true)
			{
				while(digitalCompassActive == false)
				{
					Thread.Sleep(100);
				}
				
				try
				{
					double [,] samples = getMultiSampleOnReader(readerDigitalCompass, numSamples);
			
					int len = samples.GetLength(1);
					int start = 0;
					int end = 0;

					if(writeToFile)
					{

						// escrever para ficheiro
						for(int i=0; i<len; i++)
						{
							tw.WriteLine(samples[0,i].ToString());
						}
						writeToFile = false;
					}
		

					//procurar linha baixa
					int pos = 0;
					for(; pos<len && !(samples[0,pos] < 2.5); pos++);

					// procurar inicio linha alta
					for(; pos<len && !(samples[0,pos] > 2.5); pos++);
					start = pos;

					//procurar linha baixa
					for(; pos<len && !(samples[0,pos] < 2.5); pos++);

					// procurar inicio linha alta
					for(; pos<len && !(samples[0,pos] > 2.5); pos++);
					end = pos;

					digitalCompassHeading = convertMillisecondsToHeading(((end-start) / ((double)sampleRate) * 1000));
					
					
					/*lineHigh = true;
					// esperar q a linha fique baixa
					while(lineHigh == true)
					{
						bool[] readData = digitalReader.ReadMultiSamplePortByte
						lineHigh = readData[7];
					}

					// esperar q a linha fique alta
					while(lineHigh == false)
					{
						bool[] readData = digitalReader.ReadSingleSampleMultiLine();
						if(readData[7] == true)
						{
							startTime = System.Environment.TickCount;
							lineHigh = true;
						}
					}

					// esperar q a linha fique baixa
					while(lineHigh == true)
					{
						bool[] readData = digitalReader.ReadSingleSampleMultiLine();
						lineHigh = readData[7];
					}

					// esperar q a linha fique alta
					while(lineHigh == false)
					{
						bool[] readData = digitalReader.ReadSingleSampleMultiLine();
						if(readData[7] == true)
						{
							endTime = System.Environment.TickCount;
							lineHigh = true;
						}
					}

					digitalCompassHeading = convertMillisecondsToHeading(endTime - startTime);*/

				}
				catch (DaqException ex)
				{
					MessageBox.Show(ex.Message);
					digitalCompassHeading = Double.NaN;
				}


			}
		}


		private double convertMillisecondsToHeading(double milliseconds)
		{
			return (milliseconds - 66)*10;
		}

		private double convertVoltageToAccel1(double voltage)
		{
			double deltaVoltage = accelDev1_calibValue - voltage;
			return deltaVoltage * k_voltageToAccel;
		}

		private double convertVoltageToAccel2(double voltage)
		{
			double deltaVoltage = accelDev2_calibValue - voltage;
			return deltaVoltage * k_voltageToAccel;
		}

		private double convertVoltageToHeading(double voltage)
		{
			//return - ((650 * voltage) / (voltage - 5));

			return 218*voltage - 32.7;
		}



		public void startHeadingFollower()
		{
			hfBehavior.start();
		}

		public void stopHeadingFollower()
		{
			hfBehavior.stop();
			this.stop();
		}

		public void startBumpDetect()
		{
			bdBehavior.start();
		}

		public void stopBumpDetect()
		{
			bdBehavior.stop();
			bumpCount = 0;
		}

		
		public void notifyBump()
		{
			bumpCount++;
			ThreadStart ts = new ThreadStart(notifyBumpAux);
			Thread t = new Thread(ts);
			t.Start();
			this.stop();
		}

		private void notifyBumpAux()
		{
			Thread.Sleep(400);
			bumpCount--;
			
		}

		public bool isBumped()
		{
			return (bumpCount>0);
		}

		public void setDesiredHeading(int h)
		{
			hfBehavior.setDesiredHeading(h);
		}

		public double getOdometryX()
		{
			return x_W;

		}

		public double getOdometryY()
		{
			return y_W;
		}

		public double getHeading()
		{
			return compassDev_actualValue;
		}

		public double getDigitalHeading()
		{
			return digitalCompassHeading;
		}

		public double getVx() 
		{
			return vx;
		}

		public double getVy()
		{
			return vy;
		}

		public double getLeftMotor()
		{
			return leftMotor;
		}

		public double getRightMotor()
		{
			return rightMotor;
		}

		public string getAccelDevice1() 
		{
			return accelDevice1;
		}

		public string getAccelDevice2() 
		{
			return accelDevice2;
		}

		public string getCompassDevice()
		{
			return compassDevice;
		}

		public double getAccel1ActualValue()
		{
			return accelDev1_actualValue;
		}

		public double getAccel2ActualValue()
		{
			return accelDev2_actualValue;
		}


		public void setAccel1CalibValue(double val)
		{
			accelDev1_calibValue = val;
		}

		public void setAccel2CalibValue(double val)
		{
			accelDev2_calibValue = val;
		}

		public double getAccel1CalibValue()
		{
			return accelDev1_calibValue;
		}

		public double getAccel2CalibValue()
		{
			return accelDev2_calibValue;
		}

		public double getCompassActualValue()
		{
			return compassDev_actualValue;
		}
	}



}
